@@ -116,6 +116,7 @@ class IOMixin:
     def reset(self):
         """reset container"""
         self._data = {}
+        self._units = {}
 
     # ========================================================================
     # comparisons
@@ -177,13 +178,13 @@ class IOMixin:
     def read(self, fpath, engine=DEFAULT_EXCEL_ENGINE):
         """write to the given file, guessing format from the extension"""
         fmts = {
-            ".xlsx": (self.read_excel, {"fpath": fpath, "engine": engine}),
-            ".cfg": (self.read_configobj, {"fpath": fpath}),
-            ".ini": (self.read_configobj, {"fpath": fpath}),
+            ".xlsx": (self.read_excel, {"engine": engine}),
+            ".cfg": (self.read_configobj, {}),
+            ".ini": (self.read_configobj, {}),
         }
         frootname, ext = os.path.splitext(fpath)
         func, kwargs = fmts[ext]
-        return func(**kwargs)
+        return func(fpath=fpath, **kwargs)
 
     # ========================================================================
     # EXCEL io
