@@ -7,16 +7,109 @@ import logging
 import os
 import re
 from collections import defaultdict
+from io import StringIO
 from pprint import pprint
-import pint_pandas as pintpandas
 
 import jsonschema
+import numpy as np
 import pandas as pd
-
+import pint
+import pint_pandas as pintpandas
+import pkg_resources
 import simplejson as json
 import yaml
 
 
+def yamlcat(*files):
+    """concat YAML files and returns a StringIO object
+
+    files is an iterable of posibly:
+      * StringIO object
+      * strings:
+        - YAML content (shall begin with "---")
+        - path to an existing valid YAML file
+
+    >>> f1 = StringIO('''
+    ... ---
+    ... _length: &length
+    ...   units: mm
+    ...   default: 0
+    ... key1: 5
+    ... ''')
+    >>> f2 = '''
+    ... ---
+    ... key2: 15
+    ... key3:
+    ...   <<: *length
+    ... '''
+    >>> print(yamlcat(f1, f2).getvalue())
+    ---
+    _length: &length
+      units: mm
+      default: 0
+    key1: 5
+    key2: 15
+    key3:
+      <<: *length
+    """
+    out = StringIO()
+    out.write("---")
+    for i, fh in enumerate(files):
+        if isinstance(fh, str):
+            if fh.strip().startswith("---"):
+                # YAML was passed as a text string
+                fh = StringIO(fh)
+            else:
+                fh = open(fh)  # assume we have a path to a YAML file
+        out.write(fh.read().strip().strip("---"))
+        fh.close()
+    out.seek(0)
+    return out
+
+
+def load_yaml(*files, clean=True, debug=False):
+    """concatenate and load yaml files
+    Wrapper for:
+    * yamlcat(*files)
+    * yaml.load()
+
+    if `debug` is True, return a tuple:
+        (YAML dictionnary translation, YAML concatenated file)
+    otherwise, return the YAML dictionnary translation
+
+    >>> f1 = StringIO('''
+    ... ---
+    ... _length: &length
+    ...   units: mm
+    ...   default: 0
+    ... key1: 5
+    ... ''')
+    >>> f2 = '''
+    ... ---
+    ... key2: 15
+    ... key3:
+    ...   <<: *length
+    ... key4:
+    ...   <<: *length
+    ...   default: 1
+    ... '''
+    >>> load_yaml(f1, f2) == {'key1': 5,
+    ...                       'key2': 15,
+    ...                       'key3': {'units': 'mm', 'default': 0},
+    ...                       'key4': {'units': 'mm', 'default': 1}}
+    True
+    """
+    src = yamlcat(*files)
+    specs = yaml.load(src, Loader=yaml.FullLoader)
+    if clean:
+        # clean keys beginning with "_" if required
+        specs = {k: v for k, v in specs.items() if not k.startswith("_")}
+    if debug:
+        src.seek(0)
+        return specs, src.getvalue()
+    return specs
+
+
 # use short units
 pintpandas.PintType.ureg.default_format = "~P"
 
@@ -151,8 +244,8 @@ class DataFrameSchematizer:
         self.columns_specs = {}
         self.required = []
         self._is_units = False
-        self._units = None
-
+        self._source_units = None
+        self._target_units = None
 
     def build(self):
         """build and return schema"""
@@ -172,7 +265,7 @@ class DataFrameSchematizer:
         self.add_columns(specs)
 
     def _add_columns_from_yaml(self, yamltxt):
-        specs = yaml.load(yamltxt, Loader=yaml.FullLoader)
+        specs = load_yaml(yamltxt)
         self.add_columns(specs)
 
     def add_columns_from_string(self, txt):
@@ -192,7 +285,13 @@ class DataFrameSchematizer:
             self.add_column(name=colname, **colspec)
 
     def add_column(
-        self, name, types=("integer",), unique=False, mandatory=False, units=None, **kwargs
+        self,
+        name,
+        types=("integer",),
+        unique=False,
+        mandatory=False,
+        units=None,
+        **kwargs,
     ):
         """add a column to the schema"""
         if isinstance(types, str):
@@ -216,7 +315,7 @@ class DataFrameSchematizer:
         # ---------------------------------------------------------------------
         # handle units specifications
         if units:
-            ref['units'] = units
+            ref["units"] = units
             self._is_units = True
 
         self.columns_specs[name] = ref
@@ -224,14 +323,58 @@ class DataFrameSchematizer:
     def validate_dataframe(self, df):
         """validate dataframe against self.schema()"""
         schema = self.build()
-        # ---------------------------------------------------------------------
+        intermediate_report = defaultdict(list)  # what is not raised by jsonschema
+        # =====================================================================
         # builds mult-header from dataframes if schemas are units-aware
+        # =====================================================================
         if self._is_units:
-            df.columns = pd.MultiIndex.from_tuples(zip(df.columns, df.iloc[0].fillna("")))
-            df = df.iloc[1:]
-            df = df.pint.quantify(level=-1)
-        # ---------------------------------------------------------------------
+            # -----------------------------------------------------------------
+            # work on a copy:
+            # so we can revert back to original `df` if something is failing
+            _df = df.copy()
+            # -----------------------------------------------------------------
+            # create a multi-index columns such as pint-pandas can work
+            _df.columns = pd.MultiIndex.from_tuples(
+                zip(_df.columns, _df.iloc[0].fillna(""))
+            )
+            _df = _df.iloc[1:]
+            # -----------------------------------------------------------------
+            # collect defined units for later usage
+            self._source_units = dict(
+                _df.columns.tolist()
+            )  #  {'id': '', 'distA': 'm',... }
+            for i in range(len(_df.columns)):
+                try:
+                    _df.iloc[:, i] = pd.to_numeric(_df.iloc[:, i])
+                except:
+                    pass  # probably a column with strings
+            try:
+                # insert the correct to workaround this bug:
+                # https://github.com/hgrecco/pint-pandas/issues/49
+                _df = _df.pint.quantify(level=-1)
+            except pint.errors.UndefinedUnitError as exc:
+                self._is_units = False
+                colnames = [
+                    colname
+                    for colname, unit in self._source_units.items()
+                    if unit in exc.args
+                ]
+                for colname in colnames:
+                    intermediate_report[(colname, "*")].append(
+                        "undefined units '%s'" % self._source_units[colname]
+                    )
+                    df = df[1:]
+            except Exception as exc:
+                self._is_units = False
+                intermediate_report["uncatched unit error"].extend(list(exc.args))
+            else:
+                # -------------------------------------------------------------
+                # everything was fine, commit temporary `_df` to `df`:
+                df = _df
+                del _df
+        # =====================================================================
         # first: fill empty values as requested by schema
+        # =====================================================================
         _fillnas = {
             k: schema["properties"][k]["items"].get("default")
             for k, v in schema["properties"].items()
@@ -241,20 +384,27 @@ class DataFrameSchematizer:
         }
         if fillnas:
             df = df.fillna(value=fillnas, downcast="infer")
-        # ---------------------------------------------------------------------
+        # =====================================================================
         # convert read units to schema expected units (if required)
+        # =====================================================================
         if self._is_units:
-            _target_units = {
+            self._target_units = {
                 k: schema["properties"][k].get("units", "")
                 for k in schema["properties"]
             }
-            for col, units in _target_units.items(): 
-                df[col] = df[col].pint.to(units)
+            for col, units in self._target_units.items():
+                try:
+                    df[col] = df[col].pint.to(units)
+                except KeyError:
+                    pass
+                except AttributeError:
+                    breakpoint()
             df = df.pint.dequantify()
-            self._units = dict(df.columns.tolist())  #  {'id': '', 'distA': 'm',... }
-            df.columns = df.columns.levels[0]
-        # ---------------------------------------------------------------------
+            df.columns = [t[0] for t in df.columns.tolist()]  # restrict to level 0
+
+        # =====================================================================
         # second: validate
+        # =====================================================================
         validator = jsonschema.Draft7Validator(schema)
         # df -> dict -> json -> dict to convert NaN to None
         document = json.loads(json.dumps(df.to_dict(orient="list"), ignore_nan=True))
@@ -267,7 +417,8 @@ class DataFrameSchematizer:
             else:
                 report[(col, row)].append(error.message)
                 report[(col, row)].extend([e.message for e in error.context])
-        report = dict(report)
+        report = {**dict(report), **intermediate_report}
+
         return df, len(report) == 0, report
 
 
@@ -278,33 +429,83 @@ class ValidatorMixin:
         """called by Base class __init__()"""
         self._schemas = {}
 
-    def set_schema(self, tabname, schema):
+    def quantify(self, tabname, restrict_to_units=False):
+        df = self._data[tabname].copy()
+        schema = self._schemas[tabname]
+        if not schema._is_units:
+            raise ValueError(f"tab {tabname} is not units-aware'")
+        units = schema._target_units
+        # add dummy rows
+        dummy_row = pd.DataFrame({c: [units.get(c, "")] for c in df.columns})
+        df = pd.concat((dummy_row, df))
+        if restrict_to_units:
+            df = df[[c for c in df.columns if c in units]]
+        df.columns = pd.MultiIndex.from_tuples(zip(df.columns, df.iloc[0].fillna("")))
+        df = df.iloc[1:]
+        df = df.pint.quantify(level=-1)
+        return df
+
+    def convert_to(self, tabname, units=None):
+        """convert units-aware dataframe to units
+        units can be:
+        * None: will convert to base units
+        * string: eg. 'm^3'
+        * a dict mapping columns to units
+        """
+        df = self.quantify(tabname, restrict_to_units=True)
+        if units is None:
+            df_c = df.pint.to_base_units()
+        else:
+            if isinstance(units, str):
+                units = {c: units for c in df.columns}
+            df_c = pd.DataFrame(
+                {col: df[col].pint.to(units) for col, units in units.items()}
+            )
+        return df_c
+
+    def _set_schema(self, tabname, schema):
         """assign a schema to a tab"""
         tabnames = []
+        # ---------------------------------------------------------------------
+        # generic tabname regex: collect _data tabnames
         if tabname.startswith("^") and tabname.endswith("$"):
-            # generic tabname regex: collect _data tabnames
-            # matching regex
             for data_tabname in self._data.keys():
                 if re.match(tabname, data_tabname):
                     tabnames.append(data_tabname)
+        # ---------------------------------------------------------------------
+        # general case. Only one tabname to fill
         else:
-            # not a regex. Only one tabname to fill
             tabnames = [tabname]
+        # ---------------------------------------------------------------------
+        # iterate over tabnames (usually only one if no regex-tabname supplied)
+        # to instanciate a DataFrameSchematizer and fill with a schema
         for tabname in tabnames:
-            self._schemas[tabname] = DataFrameSchematizer()  # reset schematizer
+            self._schemas[tabname] = DataFrameSchematizer()
             self._schemas[tabname].add_columns(schema)
 
-    def read_schema(self, filepath, debug=False):
+    def read_schema(self, *filepath):
         """ assign a global schema by parsing the given filepath"""
+        if len(filepath) > 1:
+            return self.read_multiple_yamls_schemas(*filepath)
+        else:
+            filepath = filepath[0]
         _, ext = os.path.splitext(filepath)
         with open(filepath, "r") as fh:
             schema_specs = fh.read()
         if ext == ".json":
             schemas = json.loads(schema_specs)
         elif ext == ".yaml":
-            schemas = yaml.load(schema_specs, Loader=yaml.FullLoader)
+            schemas = load_yaml(schema_specs)
         for tabname, schema in schemas.items():
-            self.set_schema(tabname, schema)
+            self._set_schema(tabname, schema)
+
+    def read_multiple_yamls_schemas(self, *files):
+        """
+        replace `read_schema` in case shcema is made of several YAML files
+        """
+        schemas = load_yaml(*files)
+        for tabname, schema in schemas.items():
+            self._set_schema(tabname, schema)
 
     def _validate_tab(self, tabname):
         """validate a tab using the provided scheme"""
@@ -317,7 +518,7 @@ class ValidatorMixin:
         iterate through all tabs and validate eachone
         """
         # keep initial data before processing them
-        if not hasattr(self, '_raw_data'):
+        if not hasattr(self, "_raw_data"):
             self._raw_data = {tabname: df.copy() for tabname, df in self._data.items()}
         ret = {}
         for tabname, df in self._data.items():
